"""Interfacing with Raumfeld web service"""

import xmltodict


async def async_connect_room_to_zone(session, location, zone_udn=None, room_udn=None):
    """Puts the room with the given room_udn in the zone with the zone_udn.

    Optional parameters:
    zone_udn -- The udn of the zone to connect the room to. If zone udn is
    empty or there is no zone with this udn yet, then a new zone is
    created.
    room_udn -- The udn of the room that has to be put into that zone. If
    empty, all available rooms (rooms that have active renderers) are put
    into the zone.
    """
    params = {}

    if zone_udn:
        params["zoneUDN"] = zone_udn

    if room_udn:
        params["roomUDN"] = room_udn

    url = location + "/connectRoomToZone"

    await session.get(url, params=params)


async def async_connect_rooms_to_zone(session, location, zone_udn=None, room_udns=None):
    """Puts the rooms with the given roomUDNs in the zone with the zoneUDN.

    Optional parameters:
    zone_udn -- The udn of the zone to connect the rooms to. If zone udn is
    empty or there is no zone with this udn yet, then a new zone is
    created.
    room_udns -- A list of UDNs of the rooms that have to be put into that
    zone. If empty, all available rooms (rooms that have active renderers)
    are put into the zone and activated.
    """
    params = {}

    if zone_udn:
        params["zoneUDN"] = zone_udn

    if room_udns:
        params["roomUDNs"] = ",".join(room_udns)

    url = location + "/connectRoomsToZone"

    await session.get(url, params=params)


async def async_drop_room_job(session, location, room_udn):
    """Drops the room with the given roomUDN from the zone it is in.

    Parameter:
    room_udn -- The udn of the room that has to be dropped.
    """
    params = {"roomUDN": room_udn}
    url = location + "/dropRoomJob"
    await session.get(url, params=params)


async def async_enter_automatic_standby(session, location, room_udn):
    """Calls RPC to put a room into automatic standby.

    Parameter:
    room_udn -- udn of the desired room.
    """
    params = {"roomUDN": room_udn}
    url = location + "/enterAutomaticStandby"
    await session.get(url, params=params)


async def async_enter_manual_standby(session, location, room_udn):
    """Calls RPC to put a room into manual standby.

    Parameter:
    room_udn -- udn of the desired room.
    """
    params = {"roomUDN": room_udn}
    url = location + "/enterManualStandby"
    await session.get(url, params=params)


async def async_leave_standby(session, location, room_udn):
    """Calls RPC to let a room leave manual or automatic standby.

    Parameter:
    room_udn -- udn of the desired room.
    """
    params = {"roomUDN": room_udn}
    url = location + "/leaveStandby"
    await session.get(url, params=params)


async def async_ping(session, location):
    """Just a heart beat tester

    Returns a dictionary containing hardware model and number.
    """
    url = location + "/Ping"
    response = await session.get(url)
    pong = xmltodict.parse(await response.text())

    if "response" in pong:
        return pong["response"]



"""Module to interface with Raumfeld smart speakers."""
import asyncio
import json
import sys
import threading
from time import sleep

import aiohttp
import requests
import xmltodict

from . import auxilliary as aux
from . import upnp
from . import webservice as ws
from .common import log_critical, log_debug, log_info, log_warn, logger
from .constants import (BROWSE_CHILDREN, CID_SEARCH_ALLTRACKS,
                        DEFAULT_PORT_WEBSERVICE, DELAY_FAST_UPDATE_CHECKS,
                        DELAY_REQUEST_FAILURE_LONG_POLLING, MAX_RETRIES,
                        PREFERRED_TIMEOUT_LONG_POLLING, REQUIRED_METADATA,
                        SOUND_SUCCESS, SPOTIFY_ACTIVE, TIMEOUT_LONG_POLLING,
                        TIMEOUT_WEBSERVICE_ACTION,
                        TRANSPORT_STATE_TRANSITIONING, TRIGGER_UPDATE_DEVICES,
                        TRIGGER_UPDATE_HOST_INFO, TRIGGER_UPDATE_SYSTEM_STATE,
                        TRIGGER_UPDATE_ZONE_CONFIG, TYPE_MEDIA_SERVER,
                        TYPE_RAUMFELD_DEVICE, USER_AGENT_RAUMFELD,
                        USER_AGENT_RAUMFELD_OIDS)


class RaumfeldHost:
    """Class representing a Raumfeld host"""

    callback = None

    def __init__(self, host, port=DEFAULT_PORT_WEBSERVICE, session=None):
        """Initialize raumfeld host."""
        self.host = host
        self.port = str(port)
        self.location = "http://" + self.host + ":" + self.port
        self.snap = {}
        self._loop = None

        if not session:
            log_debug("Creating session for aiohttp requests.")
            self._aiohttp_session = aiohttp.ClientSession()
        else:
            self._aiohttp_session = session
            log_debug("Session for aiohttp requests was passed.")

        # up-to-date data from Raumfeld web service
        self.wsd = {
            "devices": {},
            "host_info": {},
            "system_state": {},
            "zone_config": {},
        }

        # up-to-date data derived from "self.wsd".
        self.resolve = {
            "devudn_to_name": {},
            "room_to_udn": {},
            "udn_to_devloc": {},
            "udn_to_room": {},
            "roomudn_to_powerstate": {},
            "roomudn_to_rendudn": {},
            "zoneudn_to_roomudnlst": {},
            "zone_to_rooms": {},
        }

        self.lists = {
            "locations": [],
            "raumfeld_device_udns": [],
            "rooms": [],
            "spotify_renderer": [],
            "zones": [],
        }

        self.init_done = {
            "devices": False,
            "host_info": False,
            "system_state": False,
            "zone_config": False,
        }

        # up-to-date data derived from "self.wsd".
        self.media_server_udn = ""
        self.update_available = False

    def set_logging_level(self, level):
        """set logging level of hassfeld."""
        logger.setLevel(level)

    async def async_host_is_valid(self):
        """Check whether host is a valid raumfeld host."""
        url = self.location + "/getHostInfo"
        timeout = aiohttp.ClientTimeout(total=3)
        session = aiohttp.ClientSession(timeout=timeout)

        try:
            response = await session.get(url)
            response_xml = await response.read()
            await session.close()
        except:
            await session.close()
            return False

        host_info = xmltodict.parse(response_xml)
        return bool("hostName" in host_info["hostInfo"])

    #
    # Functions for backgorund data updates from Raumfeld host.
    #

    def start_update_thread(self):
        """Start dedicated thread for web service data updates."""
        # Background thread updating "self.wsd".
        self._loop = asyncio.new_event_loop()
        threading.Thread(target=self._loop.run_forever, daemon=True).start()

        asyncio.run_coroutine_threadsafe(self.__async_update(), self._loop)

        # Wait for first data as background updates are asynchronous.
        while False in self.init_done.values():
            sleep(DELAY_FAST_UPDATE_CHECKS)

    async def __async_update(self):
        """Execut all update loops in a group."""
        aiohttp_session = aiohttp.ClientSession()
        await asyncio.gather(
            self.async_update_gethostinfo(aiohttp_session),
            self.async_update_getzones(aiohttp_session),
            self.async_update_listdevices(aiohttp_session),
            self.async_update_systemstatechannel(aiohttp_session),
        )

    async def async_update_gethostinfo(self, session):
        """Update loop for host information."""
        while True:
            url = self.location + "/getHostInfo"
            await self.__long_polling(session, url, self.__update_host_info)

    async def async_update_getzones(self, session):
        """Update loop for zone information."""
        while True:
            url = self.location + "/getZones"
            await self.__long_polling(session, url, self.__update_zone_config)

    async def async_update_listdevices(self, session):
        """Update loop for device information."""
        while True:
            url = self.location + "/listDevices"
            await self.__long_polling(session, url, self.__update_devices)

    async def async_update_systemstatechannel(self, session):
        """Update loop for software update information."""
        while True:
            url = self.location + "/SystemStateChannel"
            await self.__long_polling(session, url, self.__update_system_state)

    async def __long_polling(self, session, url, _callback):
        """Long-polling of web service interface."""
        update_id = None
        timeout = aiohttp.ClientTimeout(total=TIMEOUT_LONG_POLLING)
        prefer_wait = "wait=" + str(PREFERRED_TIMEOUT_LONG_POLLING)
        headers = {"Prefer": prefer_wait}
        while True:
            if update_id:
                headers["updateID"] = update_id
            try:
                async with session.get(
                    url, headers=headers, timeout=timeout
                ) as response:
                    if response.status == 200:
                        update_id = response.headers["updateID"]
                        _callback(await response.read())
                    elif response.status != 304:
                        await asyncio.sleep(DELAY_REQUEST_FAILURE_LONG_POLLING)
                        continue
            except asyncio.exceptions.TimeoutError:
                log_info("Long-polling timed out")
            except:
                exc_info = "%s%s" % (sys.exc_info()[0], sys.exc_info()[1])
                log_critical("Long-polling failed with error: %s" % exc_info)
            await asyncio.sleep(DELAY_FAST_UPDATE_CHECKS)

    #
    # Callback functions for long polling.
    #

    def __update_host_info(self, content_xml):
        """Update internal data strucrure with host information."""
        gethostinfo = xmltodict.parse(content_xml)
        self.wsd["host_info"] = gethostinfo["hostInfo"]

        self.init_done["host_info"] = True

        if self.callback is not None:
            self.callback(TRIGGER_UPDATE_HOST_INFO)

    def __update_zone_config(self, content_xml):
        """Update internal data strucrure with zone information."""
        self.lists["rooms"] = []
        self.lists["zones"] = []
        self.resolve["roomudn_to_powerstate"] = {}
        self.resolve["room_to_udn"] = {}
        self.resolve["udn_to_room"] = {}
        self.resolve["zoneudn_to_roomudnlst"] = {}

        getzones = xmltodict.parse(content_xml, force_list=("zone", "room"))
        self.wsd["zone_config"] = getzones["zoneConfig"]

        if "zones" in self.wsd["zone_config"]:
            for zone_itm in self.wsd["zone_config"]["zones"]["zone"]:
                zone_rooms = []
                zone_udn = zone_itm["@udn"]
                self.resolve["zoneudn_to_roomudnlst"][zone_udn] = []

                for room_itm in zone_itm["room"]:
                    room_name = room_itm["@name"]
                    room_udn = room_itm["@udn"]
                    renderer_udn = room_itm["renderer"]["@udn"]
                    zone_rooms.append(room_name)
                    self.lists["rooms"].append(room_name)
                    if "@powerState" in room_itm:
                        self.resolve["roomudn_to_powerstate"][room_udn] = room_itm[
                            "@powerState"
                        ]
                    else:
                        log_warn(
                            "No 'powerState' attribute provided for room: %s"
                            % room_name
                        )
                        self.resolve["roomudn_to_powerstate"][room_udn] = None
                    self.resolve["room_to_udn"][room_name] = room_udn
                    self.resolve["udn_to_room"][room_udn] = room_name
                    self.resolve["zoneudn_to_roomudnlst"][zone_udn].append(room_udn)
                    self.resolve["roomudn_to_rendudn"][room_udn] = renderer_udn

                self.lists["zones"].append(sorted(zone_rooms))

        if "unassignedRooms" in self.wsd["zone_config"]:
            for room in self.wsd["zone_config"]["unassignedRooms"]["room"]:
                room_name = room["@name"]
                room_udn = room["@udn"]
                renderer_udn = room["renderer"]["@udn"]
                if "@powerState" in room:
                    self.resolve["roomudn_to_powerstate"][room_udn] = room[
                        "@powerState"
                    ]
                else:
                    log_warn(
                        "No 'powerState' attribute provided for room: %s" % room_name
                    )
                    self.resolve["roomudn_to_powerstate"][room_udn] = None
                self.resolve["room_to_udn"][room_name] = room_udn
                self.resolve["udn_to_room"][room_udn] = room_name
                self.lists["rooms"].append(room_name)
                self.resolve["roomudn_to_rendudn"][room_udn] = renderer_udn
                if "@spotifyConnect" in room["renderer"]:
                    if room["renderer"]["@spotifyConnect"] == SPOTIFY_ACTIVE:
                        self.lists["spotify_renderer"].append(renderer_udn)

        self.init_done["zone_config"] = True

        if self.callback is not None:
            self.callback(TRIGGER_UPDATE_ZONE_CONFIG)

    def __update_devices(self, content_xml):
        """Update internal data strucrure with device information."""
        self.lists["locations"] = []
        self.lists["raumfeld_device_udns"] = []
        self.resolve["devudn_to_name"] = {}
        self.resolve["udn_to_devloc"] = {}

        listdevices = xmltodict.parse(content_xml, force_list=("device"))
        self.wsd["devices"] = listdevices["devices"]["device"]

        for device_itm in self.wsd["devices"]:
            device_loc = device_itm["@location"]
            device_type = device_itm["@type"]
            device_udn = device_itm["@udn"]
            if "#text" in device_itm:
                device_name = device_itm["#text"]
            else:
                log_warn("Missing '#text' key for device with UDN: %s" % device_udn)
            self.lists["locations"].append(device_loc)
            self.resolve["devudn_to_name"][device_udn] = device_name
            self.resolve["udn_to_devloc"][device_udn] = device_loc

            if device_type == TYPE_MEDIA_SERVER:
                self.media_server_udn = device_udn

            if device_type == TYPE_RAUMFELD_DEVICE:
                self.lists["raumfeld_device_udns"].append(device_udn)

        self.init_done["devices"] = True

        if self.callback is not None:
            self.callback(TRIGGER_UPDATE_DEVICES)

    def __update_system_state(self, content_xml):
        """Update internal data strucrure with software update information."""
        systemstatechannel = xmltodict.parse(content_xml)
        self.wsd["system_state"] = systemstatechannel["systemState"]

        update_available = self.wsd["system_state"]["updateAvailable"]["@value"]

        self.update_available = aux.str_to_bool(update_available)

        self.init_done["system_state"] = True

        if self.callback is not None:
            self.callback(TRIGGER_UPDATE_SYSTEM_STATE)

    #
    #  Helper functions
    #

    def get_zones(self):
        """Get zones from RaumfeldHost class.

        Returns list of sorted lists of rooms per zone. E.g.:
        [
            ['Badezimmer DG', 'Badezimmer OG', 'Schlafzimmer', 'Werkraum'],
            ['Wohnzimmer'],
            ["Filmraum", 'Küche']
        ]
        """
        zone_lst = [sorted(x) for x in self.lists["zones"]]
        return zone_lst

    def get_rooms(self):
        """Return list of rooms."""
        room_lst = self.lists["rooms"]
        return room_lst

    def get_raumfeld_device_udns(self):
        """Return list of unified device names."""
        devudn_lst = self.lists["raumfeld_device_udns"]
        return devudn_lst

    def device_udn_to_name(self, device_udn):
        """Convert device udn to device name."""
        device_name = self.resolve["devudn_to_name"][device_udn]
        return device_name

    def device_udn_to_location(self, device_udn):
        """Convert device udn to device location."""
        device_location = self.resolve["udn_to_devloc"][device_udn]
        return device_location

    def get_host_room(self):
        """Return room of raumfeld host."""
        return self.wsd["host_info"]["roomName"]

    def get_host_name(self):
        """Return raumfeld host's host name."""
        return self.wsd["host_info"]["hostName"]

    def roomlst_to_udnlst(self, room_lst):
        """Convert list of room names to list of unique device names."""
        udn_lst = []
        for room in room_lst:
            room_udn = self.resolve["room_to_udn"][room]
            udn_lst.append(room_udn)
        return udn_lst

    def roomlst_to_zoneudn(self, room_lst):
        """Convert list of rooms to zone UDN."""
        udn_lst = self.roomlst_to_udnlst(room_lst)
        zone_udn = self.roomudnlst_to_zoneudn(udn_lst)
        return zone_udn

    def roomudnlst_to_zoneudn(self, udn_lst):
        """Convert list of room UDN to zone UDN."""
        match = None
        zone_udnlst = self.resolve["zoneudn_to_roomudnlst"].keys()
        for zone_udn in zone_udnlst:
            if zone_udn in self.resolve["zoneudn_to_roomudnlst"]:
                zoneroom_udnlst = self.resolve["zoneudn_to_roomudnlst"][zone_udn]
                if aux.lists_have_same_values(zoneroom_udnlst, udn_lst):
                    match = zone_udn
                    break
        return match

    async def __async_wait_zone_creation(self, old_zone_udn, new_zone_room_lst):
        """Wait for zone creation published and recevied."""
        max_attempts = int(TIMEOUT_WEBSERVICE_ACTION / DELAY_FAST_UPDATE_CHECKS)
        for i in range(0, max_attempts):
            new_zone_udn = self.roomudnlst_to_zoneudn(new_zone_room_lst)
            if new_zone_udn is None:
                await asyncio.sleep(DELAY_FAST_UPDATE_CHECKS)
            elif new_zone_udn == old_zone_udn:
                await asyncio.sleep(DELAY_FAST_UPDATE_CHECKS)
            elif new_zone_udn not in self.resolve["udn_to_devloc"]:
                await asyncio.sleep(DELAY_FAST_UPDATE_CHECKS)
            else:
                break

    def zone_is_valid(self, room_lst):
        """Check whether passed zone is valid."""
        zone = sorted(room_lst)
        return bool(zone in self.lists["zones"])

    def room_is_valid(self, room):
        """Check whether passed room is valid."""
        return bool(room in self.lists["rooms"])

    def rooms_are_valid(self, room_lst):
        """Check whether passed rooms are valid."""
        for room in room_lst:
            if not self.room_is_valid(room):
                return False
        return True

    def room_is_spotify_single_room(self, room):
        """Check whether passed room is in spotify single-room mode."""
        room_udn = self.resolve["room_to_udn"][room]
        if room_udn in self.resolve["roomudn_to_rendudn"]:
            rend_udn = self.resolve["roomudn_to_rendudn"][room_udn]
            return bool(rend_udn in self.lists["spotify_renderer"])
        return False

    def location_is_valid(self, location):
        """Check whether passed location is valid."""
        return bool(location in self.lists["locations"])

    def get_room_power_state(self, room):
        """Get current power state of a room."""
        if self.room_is_valid(room):
            room_udn = self.resolve["room_to_udn"][room]
            power_state = self.resolve["roomudn_to_powerstate"][room_udn]
        else:
            power_state = None
        return power_state

    #
    # Raumfeld manipulation methods
    #

    async def async_create_zone(self, room_lst):
        """Create a new zone based on list of rooms.

        Parameters:
        room_lst -- List of rooms defining a zone.
        """
        udn_lst = self.roomlst_to_udnlst(room_lst)
        old_zone_udn = self.roomudnlst_to_zoneudn(udn_lst)
        await ws.async_connect_rooms_to_zone(
            self._aiohttp_session, self.location, room_udns=udn_lst
        )
        # Zone creation may take some time before taking effect.
        await self.__async_wait_zone_creation(old_zone_udn, room_lst)

    def set_zone_room_volume(self, zone_room_lst, volume, room_lst=None):
        """Sets volume of rooms in a zone to same level."""
        return asyncio.run(
            self.async_set_zone_room_volume(zone_room_lst, volume, room_lst)
        )

    async def async_set_zone_room_volume(self, zone_room_lst, volume, room_lst=None):
        """Sets volume of rooms in a zone to same level.

        Parameters:
        zone_room_lst -- List of rooms defining a zone
        volume -- Volume to to set for rooms.

        Optional parameters:
        room_lst -- List of rooms to change volume for. If absent, defaults to
        all rooms.
        """
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)

        if room_lst is None:
            room_udnlst = self.roomlst_to_udnlst(zone_room_lst)
        else:
            room_udnlst = self.roomlst_to_udnlst(room_lst)

        if zone_udn:
            for room_udn in room_udnlst:
                zone_loc = self.resolve["udn_to_devloc"][zone_udn]
                await upnp.async_set_room_volume(
                    self._aiohttp_session, zone_loc, room_udn, volume, instance_id=0
                )

    def set_zone_mute(self, zone_room_lst, mute=True):
        """Mute zone."""
        return asyncio.run(self.async_set_zone_mute(zone_room_lst, mute))

    async def async_set_zone_mute(self, zone_room_lst, mute=True):
        """Mute zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        await upnp.async_set_mute(self._aiohttp_session, zone_loc, mute)

    def get_zone_mute(self, zone_room_lst):
        """Get mute status of zone."""
        return asyncio.run(self.async_get_zone_mute(zone_room_lst))

    async def async_get_zone_mute(self, zone_room_lst):
        """Get mute status of zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        mute = await upnp.async_get_mute(self._aiohttp_session, zone_loc)
        return mute

    def set_zone_volume(self, zone_room_lst, volume):
        """Set volume to absolute level."""
        return asyncio.run(self.async_set_zone_volume(zone_room_lst, volume))

    async def async_set_zone_volume(self, zone_room_lst, volume):
        """Set volume to absolute level."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        await upnp.async_set_volume(self._aiohttp_session, zone_loc, volume)

    def change_zone_volume(self, zone_room_lst, amount):
        """Change the volume of zone up or down."""
        return asyncio.run(self.async_change_zone_volume(zone_room_lst, amount))

    async def async_change_zone_volume(self, zone_room_lst, amount):
        """Change the volume of zone up or down."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        await upnp.async_change_volume(self._aiohttp_session, zone_loc, amount)

    def zone_stop(self, zone_room_lst):
        """Stop playing media on zone."""
        return asyncio.run(self.async_zone_stop(zone_room_lst))

    async def async_zone_stop(self, zone_room_lst):
        """Stop playing media on zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        await upnp.async_stop(self._aiohttp_session, zone_loc)

    def zone_play(self, zone_room_lst):
        """Play media on zone."""
        return asyncio.run(self.async_zone_play(zone_room_lst))

    async def async_zone_play(self, zone_room_lst):
        """Play media on zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        await upnp.async_play(self._aiohttp_session, zone_loc)

    def zone_pause(self, zone_room_lst):
        """Pause playing media on zone."""
        return asyncio.run(self.async_zone_pause(zone_room_lst))

    async def async_zone_pause(self, zone_room_lst):
        """Pause playing media on zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        await upnp.async_pause(self._aiohttp_session, zone_loc)

    def zone_seek(self, zone_room_lst, target):
        """Seek to position on zone."""
        return asyncio.run(self.async_zone_seek(zone_room_lst, target))

    async def async_zone_seek(self, zone_room_lst, target):
        """Seek to position on zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        await upnp.async_seek(self._aiohttp_session, zone_loc, "ABS_TIME", target)

    def zone_next_track(self, zone_room_lst):
        """Play next track of zone."""
        return asyncio.run(self.async_zone_next_track(zone_room_lst))

    async def async_zone_next_track(self, zone_room_lst):
        """Play next track of zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        await upnp.async_next_track(self._aiohttp_session, zone_loc)

    def zone_previous_track(self, zone_room_lst):
        """Play previous track of zone."""
        return asyncio.run(self.async_zone_previous_track(zone_room_lst))

    async def async_zone_previous_track(self, zone_room_lst):
        """Play previous track of zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        await upnp.async_previous_track(self._aiohttp_session, zone_loc)

    def browse_media_server(self, object_id, browse_flag):
        """Browse media on the media server."""
        return asyncio.run(self.async_browse_media_server(object_id, browse_flag))

    async def async_browse_media_server(self, object_id, browse_flag):
        """Browse media on the media server."""
        http_headers = None

        if browse_flag == BROWSE_CHILDREN:
            if object_id in USER_AGENT_RAUMFELD_OIDS:
                http_headers = {"User-Agent": USER_AGENT_RAUMFELD}

        media_server_loc = self.resolve["udn_to_devloc"][self.media_server_udn]
        return await upnp.async_browse(
            self._aiohttp_session,
            media_server_loc,
            object_id,
            browse_flag,
            http_headers=http_headers,
        )

    def search_media_server(
        self,
        container_id=0,
        search_criteria="",
        filter_criteria="*",
        starting_index=0,
        requested_count=0,
        sort_criteria="",
    ):
        """Search the media server."""
        return asyncio.run(
            self.async_search_media_server(
                container_id,
                search_criteria,
                filter_criteria,
                starting_index,
                requested_count,
                sort_criteria,
            )
        )

    async def async_search_media_server(
        self,
        container_id=0,
        search_criteria="",
        filter_criteria="*",
        starting_index=0,
        requested_count=0,
        sort_criteria="",
    ):
        """Search the media server.

        Parameters:
        search_criteria='raumfeld:any contains "No son of mine"
        """
        media_server_loc = self.resolve["udn_to_devloc"][self.media_server_udn]
        response = await upnp.async_search(
            self._aiohttp_session,
            media_server_loc,
            container_id,
            search_criteria,
            filter_criteria,
            starting_index,
            requested_count,
            sort_criteria,
        )
        return response

    def search_for_play(self, container_id, search_criteria):
        """Search for media and return uri and meta data."""
        return asyncio.run(self.async_search_for_play(container_id, search_criteria))

    async def async_search_for_play(self, container_id, search_criteria):
        """Search for media and return uri and meta data."""
        requested_count = 1
        sort_criteria = "+upnp:artist,-dc:date,+dc:title"
        metadata_xml = await self.async_search_media_server(
            container_id,
            search_criteria,
            requested_count=(requested_count),
            sort_criteria=sort_criteria,
        )
        metadata = xmltodict.parse(metadata_xml)

        if "item" in metadata["DIDL-Lite"]:
            uri = metadata["DIDL-Lite"]["item"]["res"]["#text"]
        else:
            uri = None

        return uri, metadata_xml

    def set_av_transport_uri(
        self, zone_room_lst, current_uri, current_uri_metadata=None
    ):
        """Set the URI of the track to play and it's meta data in a zone."""
        return asyncio.run(
            self.async_set_av_transport_uri(
                zone_room_lst, current_uri, current_uri_metadata
            )
        )

    async def async_set_av_transport_uri(
        self, zone_room_lst, current_uri, current_uri_metadata=None
    ):
        """Set the URI of the track to play and it's meta data in a zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        if current_uri_metadata is None:
            current_uri_metadata = xmltodict.unparse(REQUIRED_METADATA)
        await upnp.async_set_av_transport_uri(
            self._aiohttp_session,
            zone_loc,
            current_uri,
            current_uri_metadata,
            instance_id=0,
        )

    def search_and_zone_play(
        self, zone_room_lst, search_criteria, container_id=CID_SEARCH_ALLTRACKS
    ):
        """Search for track and play first found."""
        return asyncio.run(
            self.async_search_and_zone_play(
                zone_room_lst, search_criteria, container_id
            )
        )

    async def async_search_and_zone_play(
        self, zone_room_lst, search_criteria, container_id=CID_SEARCH_ALLTRACKS
    ):
        """Search for track and play first found."""
        uri, uri_metadata = await self.async_search_for_play(
            container_id, search_criteria
        )
        await self.async_set_av_transport_uri(zone_room_lst, uri, uri_metadata)

    def get_media_info(self, zone_room_lst):
        """Get media information of zone."""
        return asyncio.run(self.async_get_media_info(zone_room_lst))

    async def async_get_media_info(self, zone_room_lst):
        """Get media information of zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        return await upnp.async_get_media_info(self._aiohttp_session, zone_loc)

    def get_play_mode(self, zone_room_lst):
        """Get play mode of zone."""
        return asyncio.run(self.async_get_play_mode(zone_room_lst))

    async def async_get_play_mode(self, zone_room_lst):
        """Get play mode of zone."""
        transport_info = await self.async_get_transport_settings(zone_room_lst)
        return transport_info["PlayMode"]

    def get_transport_info(self, zone_room_lst):
        """Get transport information of zone."""
        return asyncio.run(self.async_get_transport_info(zone_room_lst))

    async def async_get_transport_info(self, zone_room_lst):
        """Get transport information of zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        return await upnp.async_get_transport_info(self._aiohttp_session, zone_loc)

    def get_zone_volume(self, zone_room_lst):
        """Get volume of zone."""
        return asyncio.run(self.async_get_zone_volume(zone_room_lst))

    async def async_get_zone_volume(self, zone_room_lst):
        """Get volume of zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        return await upnp.async_get_volume(self._aiohttp_session, zone_loc)

    def get_position_info(self, zone_room_lst):
        """Get play information from zone."""
        return asyncio.run(self.async_get_position_info(zone_room_lst))

    async def async_get_position_info(self, zone_room_lst):
        """Get play information from zone."""
        response = False
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        if zone_udn is not None:
            zone_loc = self.resolve["udn_to_devloc"][zone_udn]
            response = await upnp.async_get_position_info(
                self._aiohttp_session, zone_loc
            )
        return response

    def get_zone_position(self, zone_room_lst):
        """Get play position from zone."""
        return asyncio.run(self.async_get_zone_position(zone_room_lst))

    async def async_get_zone_position(self, zone_room_lst):
        """Get play position from zone."""
        position_info = await self.async_get_position_info(zone_room_lst)
        return position_info["AbsTime"]

    def get_transport_settings(self, zone_room_lst):
        """Get transport settings from zone."""
        return asyncio.run(self.async_get_transport_settings(zone_room_lst))

    async def async_get_transport_settings(self, zone_room_lst):
        """Get transport settings from zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        return await upnp.async_get_transport_settings(self._aiohttp_session, zone_loc)

    def set_play_mode(self, zone_room_lst, play_mode):
        """Set play mode of zone."""
        return asyncio.run(self.async_set_play_mode(zone_room_lst, play_mode))

    async def async_set_play_mode(self, zone_room_lst, play_mode):
        """Set play mode of zone."""
        zone_udn = self.roomlst_to_zoneudn(zone_room_lst)
        zone_loc = self.resolve["udn_to_devloc"][zone_udn]
        await upnp.async_set_play_mode(self._aiohttp_session, zone_loc, play_mode)

    def room_play_system_sound(self, room, sound=SOUND_SUCCESS):
        """Play system sound on a room."""
        return asyncio.run(self.async_room_play_system_sound(room, sound))

    async def async_room_play_system_sound(self, room, sound=SOUND_SUCCESS):
        """Play system sound on a room."""
        room_udn = self.resolve["room_to_udn"][room]
        rend_udn = self.resolve["roomudn_to_rendudn"][room_udn]
        rend_loc = self.resolve["udn_to_devloc"][rend_udn]
        await upnp.async_play_system_sound(self._aiohttp_session, rend_loc, sound)

    def save_zone(self, zone_room_lst, repl_snap=False):
        """Create backup of media state for later restore."""
        return asyncio.run(self.async_save_zone(zone_room_lst, repl_snap))

    async def async_save_zone(self, zone_room_lst, repl_snap=False):
        """Create backup of media state for later restore."""
        media_info = await self.async_get_media_info(zone_room_lst)
        position_info = await self.async_get_position_info(zone_room_lst)
        volume = await self.async_get_zone_volume(zone_room_lst)
        mute = await self.async_get_zone_mute(zone_room_lst)
        key = repr(zone_room_lst)
        if key not in self.snap or repl_snap:
            self.snap[key] = {}
            self.snap[key]["uri"] = media_info["CurrentURI"]
            self.snap[key]["metadata"] = media_info["CurrentURIMetaData"]
            self.snap[key]["abs_time"] = position_info["AbsTime"]
            self.snap[key]["volume"] = volume
            self.snap[key]["mute"] = mute

    def restore_zone(self, zone_room_lst, del_snap=True):
        """restore media state from previous snapshot."""
        return asyncio.run(self.async_restore_zone(zone_room_lst, del_snap))

    async def async_restore_zone(self, zone_room_lst, del_snap=True):
        """restore media state from previous snapshot."""
        key = repr(zone_room_lst)
        retries = 0
        if key in self.snap:
            uri = self.snap[key]["uri"]
            metadata = self.snap[key]["metadata"]
            abs_time = self.snap[key]["abs_time"]
            volume = self.snap[key]["volume"]
            mute = self.snap[key]["mute"]
            await self.async_set_zone_volume(zone_room_lst, 0)
            await self.async_set_zone_mute(zone_room_lst, mute)
            await self.async_set_av_transport_uri(zone_room_lst, uri, metadata)
            while retries < MAX_RETRIES:
                transport_info = await self.async_get_transport_info(zone_room_lst)
                transport_state = transport_info["CurrentTransportState"]
                if transport_state != TRANSPORT_STATE_TRANSITIONING:
                    break
                await asyncio.sleep(DELAY_FAST_UPDATE_CHECKS)
                retries += 1
            await self.async_zone_seek(zone_room_lst, abs_time)
            await self.async_set_zone_volume(zone_room_lst, volume)
            if del_snap:
                self.snap.pop(key, None)

    async def async_enter_automatic_standby(self, room):
        """Put room speakers into automatic stand-by."""
        room_udn = self.resolve["room_to_udn"][room]
        await ws.async_enter_automatic_standby(
            self._aiohttp_session, self.location, room_udn
        )

    async def async_enter_manual_standby(self, room):
        """Put room speakers into manual stand-by (turn off)."""
        room_udn = self.resolve["room_to_udn"][room]
        await ws.async_enter_manual_standby(
            self._aiohttp_session, self.location, room_udn
        )

    async def async_leave_standby(self, room):
        """Weak room speakers up from stand-by."""
        room_udn = self.resolve["room_to_udn"][room]
        await ws.async_leave_standby(self._aiohttp_session, self.location, room_udn)

    # Spotify single-room methods

    def room_play(self, room):
        """Play media on room."""
        return asyncio.run(self.async_room_play(room))

    async def async_room_play(self, room):
        """Play media on room."""
        room_udn = self.resolve["room_to_udn"][room]
        rend_udn = self.resolve["roomudn_to_rendudn"][room_udn]
        rend_loc = self.resolve["udn_to_devloc"][rend_udn]
        await upnp.async_play(self._aiohttp_session, rend_loc)

    def room_pause(self, room):
        """Pause media on room."""
        return asyncio.run(self.async_room_pause(room))

    async def async_room_pause(self, room):
        """Pause media on room."""
        room_udn = self.resolve["room_to_udn"][room]
        rend_udn = self.resolve["roomudn_to_rendudn"][room_udn]
        rend_loc = self.resolve["udn_to_devloc"][rend_udn]
        await upnp.async_pause(self._aiohttp_session, rend_loc)

    def get_room_transport_info(self, room):
        """Get transport information of room."""
        return asyncio.run(self.async_get_room_transport_info(room))

    async def async_get_room_transport_info(self, room):
        """Get transport information of room."""
        room_udn = self.resolve["room_to_udn"][room]
        rend_udn = self.resolve["roomudn_to_rendudn"][room_udn]
        rend_loc = self.resolve["udn_to_devloc"][rend_udn]
        return await upnp.async_get_transport_info(self._aiohttp_session, rend_loc)

    def room_next(self, room):
        """Play next track of room."""
        return asyncio.run(self.async_room_next(room))

    async def async_room_next(self, room):
        """Play next track of room."""
        room_udn = self.resolve["room_to_udn"][room]
        rend_udn = self.resolve["roomudn_to_rendudn"][room_udn]
        rend_loc = self.resolve["udn_to_devloc"][rend_udn]
        await upnp.async_next_track(self._aiohttp_session, rend_loc)

    def room_previous(self, room):
        """Play previous track of room."""
        return asyncio.run(self.async_room_previous(room))

    async def async_room_previous(self, room):
        """Play previous track of room."""
        room_udn = self.resolve["room_to_udn"][room]
        rend_udn = self.resolve["roomudn_to_rendudn"][room_udn]
        rend_loc = self.resolve["udn_to_devloc"][rend_udn]
        await upnp.async_previous_track(self._aiohttp_session, rend_loc)

    def get_room_volume(self, room):
        """Get volume of room."""
        return asyncio.run(self.async_get_room_volume(room))

    async def async_get_room_volume(self, room):
        """Get volume of room."""
        room_udn = self.resolve["room_to_udn"][room]
        rend_udn = self.resolve["roomudn_to_rendudn"][room_udn]
        rend_loc = self.resolve["udn_to_devloc"][rend_udn]
        return await upnp.async_get_volume(self._aiohttp_session, rend_loc)

    def set_room_volume(self, room, volume):
        """Set volume of room."""
        return asyncio.run(self.async_set_room_volume(room, volume))

    async def async_set_room_volume(self, room, volume):
        """Set volume of room."""
        room_udn = self.resolve["room_to_udn"][room]
        rend_udn = self.resolve["roomudn_to_rendudn"][room_udn]
        rend_loc = self.resolve["udn_to_devloc"][rend_udn]
        await upnp.async_set_volume(self._aiohttp_session, rend_loc, volume)

    # Speaker methods
    def get_device_renderer(self, udn):
        return asyncio.run(self.async_get_device_renderer(udn))

    async def async_get_device_renderer(self, udn):
        location = self.device_udn_to_location(udn)
        return await upnp.async_get_device(self._aiohttp_session, location, "renderer")

    def get_device_info(self, udn):
        return asyncio.run(self.async_get_device_info(udn))

    async def async_get_device_info(self, udn):
        location = self.device_udn_to_location(udn)
        return await upnp.async_get_info(self._aiohttp_session, location)

    def get_device_manufacturer(self, udn):
        return asyncio.run(self.async_get_device_manufacturer(udn))

    async def async_get_device_manufacturer(self, udn):
        location = self.device_udn_to_location(udn)
        return await upnp.async_get_manufacturer(self._aiohttp_session, location)

    def get_device_model_name(self, udn):
        return asyncio.run(self.async_get_device_model_name(udn))

    async def async_get_device_model_name(self, udn):
        location = self.device_udn_to_location(udn)
        return await upnp.async_get_model_name(self._aiohttp_session, location)

    def get_device_update_info(self, udn):
        return asyncio.run(self.async_get_device_update_info(udn))

    async def async_get_device_update_info(self, udn):
        location = self.device_udn_to_location(udn)
        return await upnp.async_get_update_info(self._aiohttp_session, location)

    def get_device_update_info_version(self, udn):
        return asyncio.run(self.async_get_device_update_info_version(udn))

    async def async_get_device_update_info_version(self, udn):
        response = await self.async_get_device_update_info(udn)

        if "Version" in response:
            version = response["Version"]
        else:
            version = None
        return version



"""Methods implementing UPnP requests."""
import asyncio
import sys

from async_upnp_client import UpnpFactory
from async_upnp_client.aiohttp import AiohttpRequester, AiohttpSessionRequester

from .common import log_error, log_info
from .constants import (BROWSE_CHILDREN, RESPONSE_KEY_CURRENT_MUTE,
                        RESPONSE_KEY_CURRENT_VOLUME, RESPONSE_KEY_RESULT,
                        SERVICE_AV_TRANSPORT, SERVICE_CONTENT_DIRECTORY,
                        SERVICE_ID_SETUP_SERVICE, SERVICE_RENDERING_CONTROL,
                        SOUND_FAILURE, SOUND_SUCCESS, TIMEOUT_UPNP)


def exception_handler(function):
    """Handling exceptions as decorator."""

    async def new_function(*args, **kwargs):
        name = function.__name__
        try:
            result = await function(*args, **kwargs)
            return result
        except asyncio.exceptions.TimeoutError:
            log_info("Function '%s' timed out." % name)
        except:
            exc_info = "%s%s" % (sys.exc_info()[0], sys.exc_info()[1])
            log_error("Unexpected error with %s: %s" % (name, exc_info))

    return new_function


async def get_dlna_action(location, service, action, http_headers=None, session=None):
    """Return DLNA action pased on passed parameters"""
    if session:
        requester = AiohttpSessionRequester(
            timeout=TIMEOUT_UPNP, http_headers=http_headers, session=session
        )
    else:
        requester = AiohttpRequester(timeout=TIMEOUT_UPNP, http_headers=http_headers)

    factory = UpnpFactory(requester)
    device = await factory.async_create_device(location)
    service = device.service(service)
    action = service.action(action)
    return action


@exception_handler
async def async_get_mute(session, location, channel="Master", instance_id=0):
    """Returns a bolean of the mute status of a rendering service.

    Applies to:
    Digital Media Player
    Virtual Media Player

    Parameters:
    location -- URL to the device description XML of the rendering device.
    instance_id --
    channel --
    """
    action_name = "GetMute"
    upnp_action = await get_dlna_action(
        location, SERVICE_RENDERING_CONTROL, action_name, session=session
    )
    response = await upnp_action.async_call(InstanceID=instance_id, Channel=channel)
    if RESPONSE_KEY_CURRENT_MUTE in response:
        return response[RESPONSE_KEY_CURRENT_MUTE]
    return None


@exception_handler
async def async_get_media_info(session, location, instance_id=0):
    """Return media information."""
    action_name = "GetMediaInfo"
    upnp_action = await get_dlna_action(
        location, SERVICE_AV_TRANSPORT, action_name, session=session
    )
    response = await upnp_action.async_call(InstanceID=instance_id)
    response["CurrentURIMetaData"] = upnp_action.argument(
        "CurrentURIMetaData"
    ).raw_upnp_value
    return response


@exception_handler
async def async_get_transport_info(session, location, instance_id=0):
    """Return transport information."""
    action_name = "GetTransportInfo"
    upnp_action = await get_dlna_action(
        location, SERVICE_AV_TRANSPORT, action_name, session=session
    )
    response = await upnp_action.async_call(InstanceID=instance_id)
    return response


@exception_handler
async def async_get_volume(session, location, channel="Master", instance_id=0):
    """Returns the highest volume of rooms in a zone rendering service.

    Applies to:
    Digital Media Player
    Virtual Media Player

    Parameters:
    location -- URL to the device description XML of the rendering device
    in a room.
    channel --
    instance_id --
    """
    action_name = "GetVolume"
    upnp_action = await get_dlna_action(
        location, SERVICE_RENDERING_CONTROL, action_name, session=session
    )
    response = await upnp_action.async_call(InstanceID=instance_id, Channel=channel)
    if RESPONSE_KEY_CURRENT_VOLUME in response:
        return response[RESPONSE_KEY_CURRENT_VOLUME]
    return None


@exception_handler
async def async_get_position_info(session, location, instance_id=0):
    """Return position information."""
    action_name = "GetPositionInfo"
    upnp_action = await get_dlna_action(
        location, SERVICE_AV_TRANSPORT, action_name, session=session
    )
    response = await upnp_action.async_call(InstanceID=instance_id)
    response["TrackMetaData"] = upnp_action.argument("TrackMetaData").raw_upnp_value
    return response


@exception_handler
async def async_get_transport_settings(session, location, instance_id=0):
    """Return transport settings."""
    action_name = "GetTransportSettings"
    upnp_action = await get_dlna_action(
        location, SERVICE_AV_TRANSPORT, action_name, session=session
    )
    response = await upnp_action.async_call(InstanceID=instance_id)
    return response


@exception_handler
async def async_browse(
    session,
    location,
    object_id=0,
    browse_flag=BROWSE_CHILDREN,
    filter_criteria="*",
    starting_index=0,
    requested_count=0,
    sort_criteria="",
    http_headers=None,
):
    """Browse media."""
    action_name = "Browse"
    upnp_action = await get_dlna_action(
        location,
        SERVICE_CONTENT_DIRECTORY,
        action_name,
        http_headers=http_headers,
        session=session,
    )
    response = await upnp_action.async_call(
        ObjectID=object_id,
        BrowseFlag=browse_flag,
        Filter=filter_criteria,
        StartingIndex=starting_index,
        RequestedCount=requested_count,
        SortCriteria=sort_criteria,
    )
    if RESPONSE_KEY_RESULT in response:
        response[RESPONSE_KEY_RESULT] = upnp_action.argument(
            RESPONSE_KEY_RESULT
        ).raw_upnp_value
        return response[RESPONSE_KEY_RESULT]
    return None


@exception_handler
async def async_search(
    session,
    location,
    container_id=0,
    search_criteria="",
    filter_criteria="*",
    starting_index=0,
    requested_count=0,
    sort_criteria="",
):
    """Search media."""
    action_name = "Search"
    upnp_action = await get_dlna_action(
        location, SERVICE_CONTENT_DIRECTORY, action_name, session=session
    )
    response = await upnp_action.async_call(
        ContainerID=container_id,
        SearchCriteria=search_criteria,
        Filter=filter_criteria,
        StartingIndex=starting_index,
        RequestedCount=requested_count,
        SortCriteria=sort_criteria,
    )
    if RESPONSE_KEY_RESULT in response:
        return response[RESPONSE_KEY_RESULT]
    return None


@exception_handler
async def async_set_room_volume(session, location, room, desired_volume, instance_id=0):
    """Sets the volume of a rendering service in a room.

    Applies to Virtual Media Player.

    Parameters:
    location -- URL to the device description XML of the rendering device.
    room -- The unique device number of the room (UUID). The device must be
    member of the room.
    desired_volume -- Desired volume for the room.
    instance_id --
    value of False deactivates the mute status.
    """
    action_name = "SetRoomVolume"
    upnp_action = await get_dlna_action(
        location, SERVICE_RENDERING_CONTROL, action_name, session=session
    )
    await upnp_action.async_call(
        InstanceID=instance_id, Room=room, DesiredVolume=desired_volume
    )


async def async_set_mute(
    session,
    location,
    desired_mute=True,
    channel="Master",
    instance_id=0,
):
    """Sets the mute status of a rendering service.

    Applies to:
    Digital Media Player
    Virtual Media Player

    Parameters:
    location -- URL to the device description XML of the rendering device.
    instance_id
    channel --
    desired_mute --- Bolean value of True activaes the mute status and a
    value of False deactivates the mute status.
    """
    action_name = "SetMute"
    upnp_action = await get_dlna_action(
        location, SERVICE_RENDERING_CONTROL, action_name, session=session
    )
    await upnp_action.async_call(
        InstanceID=instance_id, Channel=channel, DesiredMute=desired_mute
    )


@exception_handler
async def async_set_volume(
    session, location, desired_volume, channel="Master", instance_id=0
):
    """Returns the highest volume of rooms in a zone rendering service.

    Applies to:
    Digital Media Player
    Virtual Media Player

    Parameters:
    location -- URL to the device description XML of the rendering device.
    desired_volume -- Desired volume for the zone. Relations between rooms
    are kept.
    channel --
    instance_id --
    """
    action_name = "SetVolume"
    upnp_action = await get_dlna_action(
        location, SERVICE_RENDERING_CONTROL, action_name, session=session
    )
    await upnp_action.async_call(
        InstanceID=instance_id, Channel=channel, DesiredVolume=desired_volume
    )


async def async_change_volume(session, location, amount, instance_id=0):
    """Changes the volume of all rooms in a zone up or down.

    Applies to:
    Digital Media Player
    Virtual Media Player

    Parameters:
    location -- URL to the device description XML of the rendering device.
    amount -- Amount of volume change.
    instance_id --
    """
    action_name = "ChangeVolume"
    upnp_action = await get_dlna_action(
        location, SERVICE_RENDERING_CONTROL, action_name, session=session
    )
    await upnp_action.async_call(InstanceID=instance_id, Amount=amount)


@exception_handler
async def async_stop(session, location, instance_id=0):
    """Stop playing media."""
    action_name = "Stop"
    upnp_action = await get_dlna_action(
        location, SERVICE_AV_TRANSPORT, action_name, session=session
    )
    await upnp_action.async_call(InstanceID=instance_id)


@exception_handler
async def async_play(session, location, speed="1", instance_id=0):
    """Play media."""
    action_name = "Play"
    upnp_action = await get_dlna_action(
        location, SERVICE_AV_TRANSPORT, action_name, session=session
    )
    await upnp_action.async_call(InstanceID=instance_id, Speed=speed)


@exception_handler
async def async_pause(session, location, instance_id=0):
    """Paus playing media."""
    action_name = "Pause"
    upnp_action = await get_dlna_action(
        location, SERVICE_AV_TRANSPORT, action_name, session=session
    )
    await upnp_action.async_call(InstanceID=instance_id)


# TODO: default to unit="ABS_TIME"
@exception_handler
async def async_seek(session, location, unit, target, instance_id=0):
    """Seek to position."""
    action_name = "Seek"
    upnp_action = await get_dlna_action(
        location, SERVICE_AV_TRANSPORT, action_name, session=session
    )
    await upnp_action.async_call(InstanceID=instance_id, Unit=unit, Target=target)


@exception_handler
async def async_next_track(session, location, instance_id=0):
    """Play next track."""
    action_name = "Next"
    upnp_action = await get_dlna_action(
        location, SERVICE_AV_TRANSPORT, action_name, session=session
    )
    await upnp_action.async_call(InstanceID=instance_id)


@exception_handler
async def async_previous_track(session, location, instance_id=0):
    """Play previous track."""
    action_name = "Previous"
    upnp_action = await get_dlna_action(
        location, SERVICE_AV_TRANSPORT, action_name, session=session
    )
    await upnp_action.async_call(InstanceID=instance_id)


@exception_handler
async def async_set_av_transport_uri(
    session, location, current_uri, current_uri_meta_data="", instance_id=0
):
    """Set media to play."""
    action_name = "SetAVTransportURI"
    upnp_action = await get_dlna_action(
        location, SERVICE_AV_TRANSPORT, action_name, session=session
    )
    await upnp_action.async_call(
        InstanceID=instance_id,
        CurrentURI=current_uri,
        CurrentURIMetaData=current_uri_meta_data,
    )


@exception_handler
async def async_set_play_mode(session, location, play_mode, instance_id=0):
    """Set play mode."""
    action_name = "SetPlayMode"
    upnp_action = await get_dlna_action(
        location, SERVICE_AV_TRANSPORT, action_name, session=session
    )
    await upnp_action.async_call(InstanceID=instance_id, NewPlayMode=play_mode)


@exception_handler
async def async_get_update_info(session, location):
    """Return software update information."""
    action_name = "GetUpdateInfo"
    upnp_action = await get_dlna_action(
        location, SERVICE_ID_SETUP_SERVICE, action_name, session=session
    )
    response = await upnp_action.async_call()
    return response


@exception_handler
async def async_get_info(session, location):
    """Return softwre version."""
    action_name = "GetInfo"
    upnp_action = await get_dlna_action(
        location, SERVICE_ID_SETUP_SERVICE, action_name, session=session
    )
    response = await upnp_action.async_call()
    return response["SoftwareVersion"]


@exception_handler
async def async_get_device(session, location, service):
    """Return unique device name."""
    action_name = "GetDevice"
    upnp_action = await get_dlna_action(
        location, SERVICE_ID_SETUP_SERVICE, action_name, session=session
    )
    response = await upnp_action.async_call(Service=service)
    return response["UniqueDeviceName"]


@exception_handler
async def async_get_manufacturer(session, location):
    """Return manufacturer name."""
    requester = AiohttpSessionRequester(timeout=TIMEOUT_UPNP, session=session)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device(location)
    return device.manufacturer


@exception_handler
async def async_get_model_name(session, location):
    """Return model name."""
    requester = AiohttpSessionRequester(timeout=TIMEOUT_UPNP, session=session)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device(location)
    return device.model_name


@exception_handler
async def async_play_system_sound(
    session, location, sound=SOUND_SUCCESS, instance_id=0
):
    """Plays a one out of two available system sounds.

    Devices have to be online. The sound gets mixed into potentially played
    media.

    Applies to:
    Digital Media Player

    Parameters:
    location -- URL to the device description XML of the rendering device.
    sound -- Valid values are "Success" and "Failure".
    instance_id --
    """
    if sound != SOUND_SUCCESS:
        sound = SOUND_FAILURE

    action_name = "PlaySystemSound"
    upnp_action = await get_dlna_action(
        location, SERVICE_RENDERING_CONTROL, action_name, session=session
    )
    await upnp_action.async_call(InstanceID=instance_id, Sound=sound)



"""Constants for hassfeld"""
BROWSE_CHILDREN = "BrowseDirectChildren"
BROWSE_METADATA = "BrowseMetadata"
CID_SEARCH_ARTISTS = "0/My Music/Search/TrackArtists"
CID_SEARCH_ALBUMS = "0/My Music/Search/Albums"
CID_SEARCH_COMPOSERS = "0/My Music/Search/Composers"
CID_SEARCH_ALLTRACKS = "0/My Music/Search/AllTracks"
DEFAULT_PORT_WEBSERVICE = 47365
DELAY_REQUEST_FAILURE_LONG_POLLING = 60
DELAY_FAST_UPDATE_CHECKS = 0.1
USER_AGENT_RAUMFELD = "RaumfeldControl/3.10 RaumfeldProtocol"
USER_AGENT_RAUMFELD_OIDS = ["0/RadioTime", "0/Tidal"]
MAX_RETRIES = 100
PLAY_MODE_NORMAL = "NORMAL"
PLAY_MODE_SHUFFLE = "SHUFFLE"
PLAY_MODE_REPEAT_ONE = "REPEAT_ONE"
PLAY_MODE_REPEAT_ALL = "REPEAT_ALL"
PLAY_MODE_RANDOM = "RANDOM"
POWER_ACTIVE = "active"
POWER_STANDBY_AUTOMATIC = "AUTOMATIC_STANDBY"
POWER_STANDBY_MANUAL = "MANUAL_STANDBY"
PREFERRED_TIMEOUT_LONG_POLLING = 300
REQUIRED_METADATA = {"DIDL-Lite": {"item": {"upnp:class": "object.item.audioItem"}}}
SPOTIFY_ACTIVE = "ACTIVE"
TRANSPORT_STATE_NO_MEDIA = "NO_MEDIA_PRESENT"
TRANSPORT_STATE_PAUSED = "PAUSED_PLAYBACK"
TRANSPORT_STATE_PLAYING = "PLAYING"
TRANSPORT_STATE_STOPPED = "STOPPED"
TRANSPORT_STATE_TRANSITIONING = "TRANSITIONING"
TIMEOUT_UPNP = 15
TIMEOUT_LONG_POLLING = 330
TIMEOUT_WEBSERVICE_ACTION = 5
TRIGGER_UPDATE_DEVICES = "devices"
TRIGGER_UPDATE_HOST_INFO = "host_info"
TRIGGER_UPDATE_SYSTEM_STATE = "system_state"
TRIGGER_UPDATE_ZONE_CONFIG = "zone_config"
TYPE_MEDIA_SERVER = "urn:schemas-upnp-org:device:MediaServer:1"
TYPE_RAUMFELD_DEVICE = "urn:schemas-raumfeld-com:device:RaumfeldDevice:1"
# TODO: replace CONTENT_DIRECTORY by SERVICE_CONTENT_DIRECTORY
CONTENT_DIRECTORY = "urn:schemas-upnp-org:service:ContentDirectory:1"
SERVICE_ID_SETUP_SERVICE = "urn:schemas-raumfeld-com:service:SetupService:1"
SERVICE_AV_TRANSPORT = "urn:schemas-upnp-org:service:AVTransport:1"
SERVICE_RENDERING_CONTROL = "urn:schemas-upnp-org:service:RenderingControl:1"
SERVICE_CONTENT_DIRECTORY = "urn:schemas-upnp-org:service:ContentDirectory:1"
SOUND_SUCCESS = "Success"
SOUND_FAILURE = "Failure"
RESPONSE_KEY_CURRENT_MUTE = "CurrentMute"
RESPONSE_KEY_CURRENT_VOLUME = "CurrentVolume"
RESPONSE_KEY_RESULT = "Result"



import setuptools

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hassfeld",
    version="0.3.6-alpha1",
    description="Integration of Raumfeld into Home Assistant",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/B5r1oJ0A9G/hassfeld",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    install_requires=[
        "aiohttp",
        "async_upnp_client>=0.15.0",
        "requests",
        "xmltodict",
    ]
)



"""Functions commonly used"""
import inspect
import logging
import os

from . import __name__ as MODULE_NAME

logger = logging.getLogger(MODULE_NAME)


def log_debug(message):
    """Logging of debug information."""
    name = inspect.currentframe().f_back.f_code.co_name
    filename = inspect.currentframe().f_back.f_code.co_filename
    basename = os.path.basename(filename)
    logger.debug("%s->%s: %s", basename, name, message)


def log_info(message):
    """Logging of information."""
    logger.info(message)


def log_warn(message):
    """Logging of warnings."""
    logger.warning(message)


def log_error(message):
    """Logging of errors."""
    logger.error(message)


def log_critical(message):
    """Logging of information."""
    logger.critical(message)



"""Functions to help dealining with data"""


def str_to_bool(string):
    """Convert string to bolean"""
    return string.lower() in ["true", "1", "t", "y", "yes"]


def lists_have_same_values(lst1, lst2):
    """Compare two zones disregarding order"""
    return sorted(lst1) == sorted(lst2)
